
# ========================   Visualization   ========================================
 
#' @title vis.BarPlot
#' @description A nice barplot of a \code{table} with bar lablels and absolute/relative frequencies.
#' 
#' @param inTable The table to plot with a Barplot
#' @param percentage WHether to use percentages of absolute numbers
#' @param subtitle Whether to add an informative subtitle in the barplot
#' @param main The main title of the barplot
#' @param ylab The title in the \code{Y} axes
#' @param decimals Number of decimals to use
#' @return A nice barplot
#' 
#' @author Dimitris Kavroudakis \email{dimitris123@@gmail.com}
#' @export
#' @import plotrix graphics
#' @keywords visualization
#' @family vis
#' @examples library(goal)
#' library(plotrix)
#' cars = table( c(1,1,1,1,3,2,4,3,5,4,6,5,1,5,5,2,2,2,3,3,3,1,1,2) )
#' vis.BarPlot(inTable = cars, percentage = TRUE,  subtitle = TRUE, 
#' main = "Cars", ylab = "Cars", decimals=1)
#' 
vis.BarPlot = function(inTable, percentage=F, subtitle=TRUE, main="main", ylab="", decimals=1){
  #library(plotrix)
  if (percentage){if (ylab==""){ylab="Percentage"}}
  barpos=graphics::barplot(inTable,  main=main, ylab=ylab, names.arg=unlist(dimnames(inTable)), border="blue", density=c(80) )
  sub=paste(length(inTable),"categories  | ",sum(inTable),"cases")
  if(subtitle){graphics::mtext(sub, side=3, line=0.3, cex=0.9)}
  if (percentage){
    labels=inTable/(sum(inTable)/100)
    plotrix::boxed.labels(barpos,inTable/2, paste(format(round(labels, decimals), nsmall = decimals), "%", sep="")   )
  } else {
    labels=inTable
    plotrix::boxed.labels(barpos,inTable/2, format(round(labels, decimals), nsmall = decimals) )
  }
  graphics::box()
}
